'''
Arrange outputs so that they have all metadata combined and are ready to be plotted 
'''


import os
from os.path import join as pjoin
import re
from Bio.Seq import Seq
from Bio import SeqIO, SeqFeature
from Bio.SeqRecord import SeqRecord
import pandas as pd
pd.options.mode.chained_assignment = None
import numpy as np
import warnings
import time
import sqlite3 as sql
from collections import defaultdict
import gc
import sys
import scipy.cluster.hierarchy as shc
from gtr_utils import change_ToWorkingDirectory, make_OutputDirectory, merge_ManyFiles, multiprocesssing_Submission, generate_AssemblyId, chunks


def align_SeedGenes(df_sr):
	'''
	Use global_strand to determine the relative ordering of each region
	Return dataframe df_sr with corrected gene_position called new_pos
	'''
	## arrange genes from start to end, normalize order to strand=1
	# first get the strand orientation of the seed_gene
	df_strand = df_sr[df_sr['seed_region_id']==df_sr['locus_tag']]
	df_strand['global_strand'] = df_strand['strand']
	df_strand = df_strand[['seed_region_id','global_strand']]
	df_sr = df_sr.merge(df_strand,on='seed_region_id')
	df_sr['global_strand'] = df_sr['global_strand'].astype(int)

	# for each locus tag in a gene region, assign a gene position based on whether the global strand is 1 or -1
	df_sr['gene_position'] = df_sr[df_sr['global_strand'] == 1].groupby('seed_region_id').cumcount()+1
	df_sr['gene_position_1'] = df_sr[df_sr['global_strand'] == -1].groupby('seed_region_id').cumcount(ascending=False)+1
	df_sr['gene_position'].fillna(df_sr['gene_position_1'], inplace=True)
	df_sr = df_sr.drop(columns='gene_position_1')

	# add column that tags whether a locus_tag is the seed gene for that region
	df_sr['is_seed'] = np.where(df_sr['locus_tag']==df_sr['seed_region_id'], 1, 0)

	# extract just the positions of each region seed 
	max_seed_position = df_sr[df_sr['is_seed']==1][['seed_region_id','gene_position','is_seed']]
	max_seed_position['relative_seed_position'] = max_seed_position['gene_position']
	max_seed_position = max_seed_position.drop(columns=['gene_position','is_seed'])
	# merge with df_sr
	df_sr = df_sr.merge(max_seed_position,on='seed_region_id')
	# re-calculate seed positions relative to the global_anchor_position
	df_sr['new_pos'] = df_sr['gene_position'] - df_sr['relative_seed_position']
	return(df_sr)


def subset_SimilarityMatrix(rep_list):
	'''
	'''
	df_jac = pd.read_csv(pjoin('internal_data','seed_regions_jaccard_dist.csv'))
	df_jac = df_jac.rename(columns={df_jac.columns[0]:'seed_region_id'},inplace=False)
	df_jac = df_jac[df_jac['seed_region_id'].isin(rep_list)]
	df_jac = df_jac.drop(df_jac.columns.difference(rep_list), axis=1)
	return(df_jac)



def melt_LabelDissimilarty(df_cr, df_jac):
	'''
	'''
	df_crm = df_cr[['dbscan_label','full_index_pos','mean_dissimilarity','representative_relative_dissimilarity']]
	df_crm = pd.melt(df_crm,id_vars=['dbscan_label','full_index_pos'])
	df_crm = df_crm.rename(columns={'variable':'measurement', 'value':'dissimilarity'})
	df_crm = df_crm.merge(df_jac, on ='dbscan_label')
	df_crm.to_csv(pjoin('internal_data','rtable_label_dissimilarity.csv'),index=False)


def melt_IdentityCoverage(df_sr):
	'''
	'''
	df_id = df_sr[['region_id','pident','qcovs','dbscan_label','label_representative','full_index_pos','order']].drop_duplicates()
	df_id['representative_region'] = np.where(df_id['label_representative']==df_id['full_index_pos'], 'representative','member')
	df_id = df_id[['region_id','pident','qcovs','dbscan_label','representative_region','order']]
	df_id = pd.melt(df_id, id_vars = ['region_id','dbscan_label','order','representative_region'])
	df_id.to_csv(pjoin('internal_data','rtable_label_identitycoverage.csv'),index=False)


def sort_DataFrameByHierarchicalClustering(df_input):
	'''
	Sort dataframe by hierarchical clustering. Same order as visualization outputs
	df_horder contains the hierarchical ordering output. 
	'''
	df_horder = pd.read_csv(pjoin(os.getcwd(),'internal_data','full_region.csv'))[['dbscan_label','order']].drop_duplicates()
	df_horder['order'] = np.where(df_horder['dbscan_label']==-1,-1,df_horder['order'])
	df_input = df_input.merge(df_horder,on='dbscan_label')
	df_input = df_input.sort_values(by='order', ascending=False)
	return(df_input)

def rename_ColumnsForUser(df):
	'''
	'''
	rename_dict = {'dbscan_label':'group','order':'hierarchical_order','region_id':'member_id','pseudo_check':'is_pseudogene','pident':'identity_to_query','qcovs':'coverage_to_query','evalue':'evalue_to_query',
	'mean_dissimilarity':'pairwise_mean_dissimilarity','representative_relative_dissimilarity':'relative_dissimilarity_to_representative','relative_seed_position':'relative_gene_position'}
	df = df.rename(columns=rename_dict)
	return(df)

## --------------Start script---------------------##

def OutputRegionResults(
	UserInput_main_dir,
	UserInput_output_dir_name,
	):
	'''
	'''

### Troubleshooting inputs ###
# UserInput_main_dir = '/Users/owlex/Dropbox/Documents/Northwestern/Hartmann_Lab/syntenease_project/gtr/testbed/dataset4/test1'
# UserInput_output_dir_name = 'pdua'
# pd.set_option('display.max_columns', None)

###--------------------###


	change_ToWorkingDirectory(directory_name = pjoin(UserInput_main_dir,UserInput_output_dir_name))
	conn = sql.connect('seed_results.db')
	conn2 = sql.connect(pjoin(UserInput_main_dir,'genomes.db'))

	## prepare taxonomy data ##
	df_meta = pd.read_sql_query("SELECT * FROM gb_metadata", conn2)
	df_meta['genus'] = [t.split(' ')[0].lower() for t in df_meta['organism'].tolist()]
	df_meta['species'] = [t.split(' ')[1].lower() for t in df_meta['organism'].tolist()]
	df_meta = df_meta[['assembly_id','genus','species','organism']]

	## add ortholog, mmseqs cluster ids, dbscan labels, and filtered hit results to the extracted seed regions ##
	df_sr = pd.read_sql_query("SELECT * FROM seed_regions", conn) 
	df_cr = pd.read_sql_query("SELECT * FROM dbscan_label_representatives", conn)
	df_or = pd.read_sql_query("SELECT * FROM select_mcl_results", conn)

	# ----  Make comprehensive region & cluster data table ---- #

	# merge dbscan labels with seed regions
	df_cr = df_cr[['seed_region_id','dbscan_label','mean_dissimilarity','representative_relative_dissimilarity','label_representative','full_index_pos']]
	df_sr = df_sr.merge(df_cr, left_on='region_id', right_on='seed_region_id')

	# merge ortholog clusters with seed regions. Requires doing a small recalculation as in cluster_Regions.py
	# this is becasue not all genes in seed regions have a sequence because they are listed as pseudo genes.
	df_sr = df_sr.merge(df_or, left_on = 'locus_tag',  right_on =  'orf_id', how = 'outer')
	df_sr['orf_id'] = df_sr['orf_id'].fillna(df_sr['locus_tag'])
	max_ortho_cluster_id_val = df_sr['ortho_cluster_id'].max(skipna = True)
	max_mmseqs_cluster_id_val = df_sr['mmseqs_cluster_id'].max(skipna = True)
	df_sr['ortho_cluster_id'] = df_sr['ortho_cluster_id'].fillna(max_ortho_cluster_id_val +1)
	df_sr['mmseqs_cluster_id'] = df_sr['mmseqs_cluster_id'].fillna(max_mmseqs_cluster_id_val +1)

	# add taxonomy data
	df_sr = df_sr.merge(df_meta,on='assembly_id')

	## big merge: gene, product names, for all seed_region members
	df_genomes = pd.read_sql_query("SELECT locus_tag, refseq_locus_tag, refseq_product, refseq_gene, cds_start, cds_end FROM gb_features WHERE locus_tag IN {}".format(tuple(df_sr.locus_tag.tolist())), conn2) 
	df_sr = df_sr.merge(df_genomes,on='locus_tag')
	df_sr = align_SeedGenes(df_sr=df_sr)

	## Use hierarchical clustering to order the output
	# filter for columns and rows that are representatives
	rep_list = df_cr[df_cr['full_index_pos']==df_cr['label_representative']]['seed_region_id'].tolist()
	label_list = df_cr[df_cr['full_index_pos']==df_cr['label_representative']]['dbscan_label'].tolist()
	df_jac = subset_SimilarityMatrix(rep_list=rep_list)

	# make dendrogram with column names as labels
	dend = shc.dendrogram(shc.linkage(y= df_jac, method='ward',
		optimal_ordering = True), labels = df_jac.columns)

	# make the dendrogram to a dataframe
	rep_label_map = pd.DataFrame({'representative_id':rep_list,'dbscan_label':label_list})
	df_reorder = pd.DataFrame({
		'representative_id': dend['ivl'],
		'order': range(0,len(dend['ivl']))})
	df_reorder = df_reorder.merge(rep_label_map, on = 'representative_id')
	df_sr = df_sr.merge(df_reorder.drop(columns='representative_id'), on='dbscan_label')


	# ---- Write comprehensive region & cluster data table for internal use ---- #
	df_sr.to_csv(pjoin('internal_data','full_region.csv'),index=False)

	# ---- Make metadata for plotting outputs ---- #
	melt_LabelDissimilarty(df_cr=df_cr, df_jac=df_reorder)
	melt_IdentityCoverage(df_sr=df_sr)
	df_meta.to_csv(pjoin('internal_data','rtable_genome_input_metadata.csv'),index=False)
	df_sr[['assembly_id','region_id','dbscan_label']].drop_duplicates().to_csv(pjoin('internal_data','rtable_summarized_region_clusters.csv'), index=False)

	# ---- Write summary tabular outputs for user ---- #

	## groups labels with dissimilarity, pident, qcovs info
	df_t1 = pd.read_csv(pjoin('internal_data','full_region.csv'))[['dbscan_label','region_id','mean_dissimilarity','pident','qcovs']].drop_duplicates().melt(id_vars=['dbscan_label','region_id'])
	max_result_vals = df_t1.groupby(['dbscan_label','variable']).max().reset_index().drop(columns='region_id')
	max_result_vals['variable'] = 'max_'+max_result_vals['variable']
	max_result_vals = max_result_vals.pivot_table(index='dbscan_label',columns='variable',values='value').reset_index()
	min_result_vals = df_t1.groupby(['dbscan_label','variable']).min().reset_index().drop(columns='region_id')
	min_result_vals['variable'] = 'min_'+min_result_vals['variable']
	min_result_vals = min_result_vals.pivot_table(index='dbscan_label',columns='variable',values='value').reset_index()
	mean_result_vals = df_t1.groupby(['dbscan_label','variable']).mean().reset_index()
	mean_result_vals['variable'] = 'mean_'+mean_result_vals['variable']
	mean_result_vals = mean_result_vals.pivot_table(index='dbscan_label',columns='variable',values='value').reset_index()
	sd_result_vals = df_t1.groupby(['dbscan_label','variable']).std().reset_index()
	sd_result_vals['variable'] = 'sd_'+sd_result_vals['variable']
	sd_result_vals = sd_result_vals.pivot_table(index='dbscan_label',columns='variable',values='value').reset_index()
	group_counts = pd.read_csv(pjoin('internal_data','full_region.csv'))[['dbscan_label','region_id']].drop_duplicates().groupby('dbscan_label').count().rename(columns={'region_id':'member_counts'}).reset_index()
	df_r1 = max_result_vals
	df_r1 = pd.concat([df_r1, min_result_vals.iloc[:,1:],mean_result_vals.iloc[:,1:],sd_result_vals.iloc[:,1:], group_counts.iloc[:,1]], axis = 1)
	df_r1 = df_r1[['dbscan_label','member_counts',
	'min_mean_dissimilarity','mean_mean_dissimilarity','max_mean_dissimilarity','sd_mean_dissimilarity',
	'min_pident','mean_pident','max_pident','sd_pident',
	'min_qcovs','mean_qcovs','max_qcovs','sd_qcovs']]
	df_r1 = sort_DataFrameByHierarchicalClustering(df_input=df_r1)
	df_r1 = rename_ColumnsForUser(df=df_r1)
	df_r1.to_csv('group_statistics_summmary.csv',index=None)


	## gene content for representative groups
	df_t1 = pd.read_csv(pjoin('internal_data','full_region.csv'))[['dbscan_label','region_id','assembly_id',
		'label_representative','full_index_pos','refseq_locus_tag','refseq_gene','refseq_product','gene_position','is_seed']]
	df_t1 = df_t1[df_t1['full_index_pos']==df_t1['label_representative']]
	df_t1 = df_t1[['dbscan_label','refseq_locus_tag','refseq_gene','refseq_product','gene_position','is_seed']]
	df_t1 = df_t1.replace(np.nan, 'NA', regex=True)
	df_t1['gene_info'] = df_t1['refseq_gene'] + ' | ' + df_t1['refseq_locus_tag'] + ' | ' + df_t1['refseq_product']
	df_t1['gene_info'] = np.where(df_t1['is_seed'] ==1, '* '+df_t1['gene_info'], df_t1['gene_info'])
	df_t1 = df_t1[['gene_position','dbscan_label','gene_info']]
	df_t1['dbscan_label'] = 'group_'+df_t1['dbscan_label'].astype(int).astype(str)
	df_t1 = df_t1[['gene_position','dbscan_label','gene_info']].pivot(index='gene_position',columns='dbscan_label',values='gene_info').reset_index()
	# reorder to hierarchical clustering
	df_horder = pd.read_csv(pjoin(os.getcwd(),'internal_data','full_region.csv'))[['dbscan_label','order']].drop_duplicates()
	df_horder['order'] = np.where(df_horder['dbscan_label']==-1,-1,df_horder['order'])
	df_horder = df_horder.sort_values(by='order', ascending=False)
	group_label_list = ['gene_position']
	for c in df_horder.iterrows():
		group_label_list.append('group_'+c[1][0].astype(int).astype(str))
	df_t1 = df_t1[group_label_list]
	df_t1.to_csv('representative_group_member_summary.csv',index=None)


	## group distribution by genus
	df_g1 = pd.read_csv(pjoin('internal_data','full_region.csv'))[['dbscan_label','assembly_id','genus']].drop_duplicates()
	df_g1 = df_g1.groupby(['dbscan_label','genus']).count().reset_index()
	df_g2 = pd.read_csv(pjoin('internal_data','full_region.csv'))[['dbscan_label','seed_region_id','genus']].drop_duplicates()
	df_g2 = df_g2.groupby(['dbscan_label','genus']).count().reset_index()
	df_g1 = df_g1.merge(df_g2, on = ['dbscan_label','genus'])
	df_t2 = df_meta[['assembly_id','genus']].groupby('genus').count().reset_index().rename(columns={'assembly_id':'genus_total'})
	df_g1 = df_g1.merge(df_t2, on='genus')
	df_g1['perc_present'] = ((df_g1['assembly_id']/df_g1['genus_total'])*100).round(2).astype(str)
	df_g1['perc_present'] = np.where(df_g1['assembly_id']<df_g1['seed_region_id'],df_g1['perc_present']+'*',df_g1['perc_present'])
	df_g1 = df_g1[['genus','perc_present','dbscan_label']]
	df_g1 = df_g1.pivot(index='dbscan_label',columns='genus',values='perc_present').reset_index()
	df_g1 = sort_DataFrameByHierarchicalClustering(df_input=df_g1)
	df_g1 = rename_ColumnsForUser(df=df_g1)
	df_g1.to_csv('group_taxa_summary.csv',index=None)


	## long form datasets for user
	df_l1 = pd.read_csv(pjoin('internal_data','full_region.csv'))
	df_l1 = df_l1[df_l1['label_representative']==df_l1['full_index_pos']][['dbscan_label','region_id']].drop_duplicates().rename(columns={'region_id':'group_representative'})
	df_l1 = df_l1.merge(pd.read_csv(pjoin('internal_data','full_region.csv')), on = 'dbscan_label')
	df_l1 = df_l1[['region_id','dbscan_label','group_representative',
		'assembly_id','contig_id','gene_position','orf_id','pseudo_check','refseq_locus_tag','refseq_gene','refseq_product',
		'is_seed','pident','qcovs','evalue',
		'ortho_cluster_id','ortho_cluster_members','mmseqs_cluster_id','mmseqs_cluster_representative',
		'mean_dissimilarity','representative_relative_dissimilarity',
		'organism','genus','species',
		'cds_start','cds_end','strand','relative_seed_position',
		'order']]
	df_l1 = rename_ColumnsForUser(df=df_l1)
	df_l1['group'] = df_l1['group'].astype(int)
	df_l1.to_csv('group_regions.csv',index=None)


	# close SQL connections
	conn.close()
	conn2.close()

	# clean up
	os.system('cp ./ortholog_clusters/merged_aa.faa ./group_region_seqs.faa')
	os.system('rm -R ./ortholog_clusters')
	os.system('rm -R ./regions')

	print('\nAll outputs files written to {}\n'.format(os.getcwd()))


## --------------End script---------------------##













import os
from os.path import join as pjoin
import re
import multiprocessing as mp
from multiprocessing import Pool
from Bio.Seq import Seq
from Bio import SeqIO, SeqFeature
from Bio.SeqRecord import SeqRecord
import pandas as pd
pd.options.mode.chained_assignment = None
import numpy as np
import warnings
import time
import sqlite3 as sql
from collections import defaultdict
import gc
import sys
from gtr_utils import change_ToWorkingDirectory, make_OutputDirectory, merge_ManyFiles, multiprocesssing_Submission, generate_AssemblyId, chunks
import subprocess
from subprocess import DEVNULL, STDOUT, check_call

def count_OrfSeqs(input_fasta_file):
	'''
	returns the total number of sequences in a fasta. starts at 1. Used to create an index for the large merged_aa fasta file.
	'''
	number_of_seqs = 0
	for record in SeqIO.parse(input_fasta_file,'fasta'):
		number_of_seqs += 1
	return(number_of_seqs)


def ava_Blast(query,db_name,output_filename):
	'''
	'''
	# os.system('blastp -query {} -db {} -max_target_seqs 100 -evalue 1e-6 -outfmt "10 qseqid sseqid mismatch positive gaps ppos pident qcovs evalue bitscore qframe sframe sstart send slen qstart qend qlen" -num_threads 1  -out {}'.format(query,db_name,output_filename))
	check_call(['blastp', '-query', query, '-db', db_name, '-max_target_seqs', '100', '-evalue', '1e-6', '-outfmt', "10 qseqid sseqid mismatch positive gaps ppos pident qcovs evalue bitscore qframe sframe sstart send slen qstart qend qlen", '-num_threads', '1', '-out', output_filename], stdout=DEVNULL, stderr=STDOUT)


def chunk_BlastInput(chunksize,fasta_to_chunk):
	'''
	chunk up the merged_aa file to form several query fasta files
	The input fasta file can either be a unclustered or preclustered orf fasta
	'''
	fasta_to_chunk_input = '{}_DB_clu_rep.fasta'.format(fasta_to_chunk)
	total_aa_seqs = count_OrfSeqs(input_fasta_file = fasta_to_chunk_input)


	## populate a list with the number of chunks and the remaining chunk value.
	## calculate the cumulative value for each row, this forms an index that is equivalent to the number from enumerate
	## note: -1 is necessary on the final input to match enumerate counter
	aa_chunk_splitsize = int(total_aa_seqs/chunksize)

	aa_chunk_size_index = [aa_chunk_splitsize for x in range(chunksize)]
	aa_chunk_size_index.append(total_aa_seqs-sum(aa_chunk_size_index)-1)
	if sum(aa_chunk_size_index) != total_aa_seqs-1:
		print('BAD MESSAGE: chunking error')
	aa_chunk_size_index = pd.Series(aa_chunk_size_index).cumsum().tolist()

	## chunk_tracker is updated and temp_sequences is cleared everytime the chunk size is reached and a file with the data is written
	chunk_filesuffix = 'aa_chunk_'
	chunk_tracker = 0
	temp_sequences = []
	for e, record in enumerate(SeqIO.parse(fasta_to_chunk_input,'fasta')):
		if e == aa_chunk_size_index[0]:
			temp_sequences.append('>'+record.id+'\n')
			temp_sequences.append(str(record.seq)+'\n')
			with open('{}{}.txt'.format(chunk_filesuffix, chunk_tracker),'w') as outfile:
				[outfile.write(i) for i in temp_sequences]
			temp_sequences = []
			chunk_tracker += 1
			aa_chunk_size_index = aa_chunk_size_index[1:]
		else:
			temp_sequences.append('>'+record.id+'\n')
			temp_sequences.append(str(record.seq)+'\n')


def blast_ChunkedSeqs(chunksize, fasta_to_chunk_prefix,working_directory,UserInput_processes):
	'''
	Returns tabular blast results for a chunked input blast search
	'''
	chunk_BlastInput(chunksize=chunksize,fasta_to_chunk=fasta_to_chunk_prefix)
	query_chunks = [[f,fasta_to_chunk_prefix,f.replace('.txt','.csv')] for f in os.listdir(working_directory) if f.startswith('aa_chunk_') if f.endswith('.txt')]
	multiprocesssing_Submission(function_name=ava_Blast,submission_list=query_chunks,processor_n=UserInput_processes,output_type='file')
	merged_ava_output = '{}_avablast'.format(fasta_to_chunk_prefix)
	merge_ManyFiles(input_filepath=working_directory,output_filepath=working_directory,wildcard_search='aa_chunk_*.csv',output_filename=merged_ava_output+'.csv')


def cluster_MCLOrtho(df_mcl, similarity_scoring_method, inflation_values_list):
	'''
	Convert a tabular blast result into a .abc formatted dataframe that is then clustered using mcl at different inflation values. save a csv of cluster relationship at each inflation value
	'''
	# mcl_file_prefix='merged_aa'
	df_mcl = df_mcl[['qseqid','sseqid',similarity_scoring_method]]
	df_mcl.to_csv('merged_aa.abc',sep='\t',index=None,header=None)
	#
	# os.system("mcxload -abc {}.abc --stream-mirror --stream-neg-log10  -o seq_{}.mci -write-tab seq_{}.tab -stream-tf 'ceil(200)'".format(mcl_file_prefix,mcl_file_prefix,mcl_file_prefix))
	check_call( ["mcxload", "-abc", "merged_aa.abc", "--stream-mirror", "--stream-neg-log10",  "-o", "seq_merged_aa.mci", "-write-tab", "seq_merged_aa.tab", "-stream-tf", 'ceil(200)'], stdout=DEVNULL, stderr=STDOUT)
	#
	for inval in inflation_values_list:
		## find clusters for each inflation value
		i_val = inval.replace('.','')
		# os.system('mcl seq_{}.mci -I {}'.format(mcl_file_prefix,inval))
		check_call( ['mcl', 'seq_merged_aa.mci', '-I', str(inval)], stdout=DEVNULL, stderr=STDOUT)

		# os.system('mcxdump -icl out.seq_{}.mci.I{} -tabr seq_{}.tab -o dump.seq_{}.mci.I{}'.format(mcl_file_prefix,i_val,mcl_file_prefix,mcl_file_prefix,i_val))
		check_call( ['mcxdump', '-icl', 'out.seq_merged_aa.mci.I{}'.format(i_val), '-tabr', 'seq_merged_aa.tab', '-o', 'dump.seq_merged_aa.mci.I{}'.format(i_val) ], stdout=DEVNULL, stderr=STDOUT)
		## extract clusters into csv format
		## read txt file which has members of a cluster per line
		keeps = []
		with open('dump.seq_merged_aa.mci.I{}'.format(i_val),'r') as infile:
			for line in infile:
				keeps.append(line.replace('\n','').split('\t'))
		## place in dataframe
		df_clong = pd.DataFrame()
		for e,i in enumerate(keeps):
			cl_id = [e for x in range(len(i))]
			z = pd.DataFrame({'ortho_cluster_id':cl_id,'ortho_cluster_members':i})
			df_clong = pd.concat([df_clong,z],axis = 0)
		df_clong.to_csv('mcl_i{}.csv'.format(i_val),index=None)	



def add_UnclusteredOrfs(df_unclustered,df_infval_temp,inf_val):
	'''
	Not all ORFs may end up being found during AVA blast and therefore not part of the mcl results. Adding them back in here. 
	Each orf added in is naturally its own cluster representative
	'''
	new_cluster_range_lower = df_infval_temp['ortho_cluster_id'].max()+1
	new_cluster_range_upper = len(df_unclustered) + new_cluster_range_lower
	new_clusters = [i for i in range(new_cluster_range_lower,new_cluster_range_upper)]
	df_unclustered['ortho_cluster_id']=new_clusters
	df_unclustered['i_val'] = inf_val
	df_unclustered = df_unclustered[['ortho_cluster_id','ortho_cluster_members','i_val']]
	return(df_unclustered)



def assign_ClusterMembership(df_unclustered, inflation_values_list,):
	'''
	Each orf should have a cluster assignment, whether through MCL or manually by adding in unclustered orfs using add_UnclusteredOrfs
	Output is a combined mcl cluster assignments at the different inflation values in a long df format
	'''
	df_infval = pd.DataFrame()
	for inf_val in inflation_values_list:
		i_val = inf_val.replace('.','')
		df_infval_temp = pd.read_csv('mcl_i{}.csv'.format(i_val))
		df_infval_temp['i_val'] = inf_val
		if inf_val == inflation_values_list[0]:
			## the first entry to populate df_infval needs special attention
			df_unclustered = add_UnclusteredOrfs(df_unclustered=df_unclustered,df_infval_temp=df_infval_temp,inf_val=inf_val)
			df_infval_temp = pd.concat([df_infval_temp,df_unclustered])
			df_infval = df_infval_temp
			continue
		df_unclustered = add_UnclusteredOrfs(df_unclustered=df_unclustered,df_infval_temp=df_infval_temp,inf_val=inf_val)
		## do two merges: first to get the clustered and unclustered orfs into the same df, and then to add the temp orf cluster df to the master orf cluster df
		df_infval_temp = pd.concat([df_infval_temp,df_unclustered])
		df_infval = pd.concat([df_infval,df_infval_temp])
	df_infval['ortho_cluster_id'] = df_infval['ortho_cluster_id'].astype(int)
	return(df_infval)


def filter_MCLOorthos(df, UserInput_ortho_select_method):
	'''
	User selects the method to choose which MCL inflation value is chosen
	Returns a float 
		max_div > Returns a float that is t lowest inflation value with the highest ortholog group diversity.
		lowest_i > retruns a float of the lowest i_val used
	'''
	if UserInput_ortho_select_method == 'max_div':
		ortho_diversity_array = pd.DataFrame(df[['i_val','ortho_cluster_id']].groupby('i_val')['ortho_cluster_id'].nunique()).reset_index()
		max_ortho_div_val = ortho_diversity_array['ortho_cluster_id'].max()
		ortho_diversity_array['i_val'] = ortho_diversity_array['i_val'].astype(float)
		selected_ortho_mcl_i_val = ortho_diversity_array[ortho_diversity_array['ortho_cluster_id']==max_ortho_div_val].sort_values('i_val',ascending=True).nsmallest(1,'i_val')['i_val'].item()
	if UserInput_ortho_select_method == 'lowest_i':
		selected_ortho_mcl_i_val = 1.2
	return(selected_ortho_mcl_i_val)



## --------------Start script---------------------##

def ClusterRegionSeqs(
	UserInput_main_dir,
	UserInput_output_dir_name,
	UserInput_faa_chunk_size,
	UserInput_blast_filter_params,
	UserInput_inflation_values,
	UserInput_similarity_scoring_method,
	UserInput_linclust_settings,
	UserInput_ortho_select_method,
	UserInput_processes
	):


	## make an mmseqs database ##
	change_ToWorkingDirectory(directory_name=pjoin(UserInput_main_dir,UserInput_output_dir_name,'ortholog_clusters'))
	make_OutputDirectory(new_directory='tmp_region_orfs')
	os.system('mmseqs createdb merged_aa.faa merged_aa_DB -v 2')
	os.system('mmseqs linclust merged_aa_DB merged_aa_DB_clu tmp_region_orfs {} -v 2'.format(UserInput_linclust_settings))
	os.system('mmseqs createsubdb merged_aa_DB_clu merged_aa_DB merged_aa_DB_clu_rep -v 2')
	os.system('mmseqs convert2fasta merged_aa_DB_clu_rep merged_aa_DB_clu_rep.fasta -v 2')
	os.system('mmseqs createtsv merged_aa_DB merged_aa_DB merged_aa_DB_clu merged_aa_resultsDB_clu.tsv -v 2')

	## run an all-vs-all blast on the reduced amino acid sequence set ##
	# os.system('makeblastdb -in merged_aa_DB_clu_rep.fasta -out merged_aa -dbtype prot -title "ava_db" -parse_seqids')
	check_call(['makeblastdb', '-in', 'merged_aa_DB_clu_rep.fasta', '-out', 'merged_aa', '-dbtype' ,'prot', '-title', "ava_db", '-parse_seqids'], stdout=DEVNULL, stderr=STDOUT)
	blast_ChunkedSeqs(chunksize=UserInput_faa_chunk_size, fasta_to_chunk_prefix='merged_aa',working_directory=os.getcwd(),UserInput_processes=UserInput_processes)
	df_ava = pd.read_csv('merged_aa_avablast.csv',names=['qseqid','sseqid','mismatch', 'positive','gaps', 'ppos','pident','qcovs','evalue','bitscore','qframe','sframe','sstart', 'send', 'slen', 'qstart', 'qend', 'qlen'])
	df_ava = df_ava[ (df_ava['pident'] >= UserInput_blast_filter_params[0]) & (df_ava['qcovs'] >= UserInput_blast_filter_params[1] ) ]

	## before inputting into MCL, need to make sure that all representative sequences have are present in the ava hits table ##
	## This section creates df_interseciton_orfs that contains sequences not picked up by blast and will therefore not be used for MCL ##
	## Those sequences that did not were not a part of the avablast results will be added in after MCL clustering  as their own unique ortholog cluster ##
	df_mmseqsclust = pd.read_csv('merged_aa_resultsDB_clu.tsv',sep='\t',header=None,names=['mmseqs_cluster_representative','orf_id'])
	all_orfs = set(df_mmseqsclust['mmseqs_cluster_representative'].unique().tolist()) #each locus_tag should have at least one mmseqs_cluster_representative
	avablast_orfs = set(df_ava['qseqid'].unique().tolist())
	intersection_orfs = list(all_orfs.difference(avablast_orfs))
	df_intersection_orfs = pd.DataFrame({'ortho_cluster_members':intersection_orfs})
	print('{} sequences not in AVA blast output'.format(len(df_intersection_orfs)))


	## Cluster ava blast results using MCL ##
	cluster_MCLOrtho(df_mcl=df_ava, similarity_scoring_method=UserInput_similarity_scoring_method, inflation_values_list=UserInput_inflation_values)

	## Assign cluster membership: Use MCL results for all sequences that were present in the AVA results ##
	## add missing sequences (if they exist), as their own cluster ID to the results ##
	## Store the results in df_infval, which is the clsuter assignment for each representative sequence, for each inflation value tested
	## df_infval is a long-form dataframe
	df_infval = assign_ClusterMembership(df_unclustered = df_intersection_orfs, inflation_values_list=UserInput_inflation_values)

	## Generate final output final containing mmseqs cluster representative, ortholog cluster representative, over different inflation values, for ALL sequences in merged_aa.faa ##
	df_mmseqsclust['mmseqs_cluster_id'] = df_mmseqsclust.groupby('mmseqs_cluster_representative').ngroup()
	df_infval = df_infval.merge(df_mmseqsclust, left_on='ortho_cluster_members', right_on='mmseqs_cluster_representative')


	## Store full results: All inflation value relationships
	change_ToWorkingDirectory(directory_name = pjoin(UserInput_main_dir,UserInput_output_dir_name))
	conn = sql.connect('seed_results.db')
	df_infval.to_sql(name='full_mcl_results',con=conn, if_exists='replace',index_label='locus_tag',index=False)
	df_infval.to_csv(pjoin('internal_data','full_mcl_results.csv'),index=False)

	## Store the "best" results: currently the max diversity
	selected_i_val = filter_MCLOorthos(df=df_infval, UserInput_ortho_select_method=UserInput_ortho_select_method) 
	df_infval = df_infval[df_infval['i_val'] == str(selected_i_val)]
	df_infval = df_infval.groupby('orf_id').first().reset_index()
	df_infval.to_sql(name='select_mcl_results',con=conn, if_exists='replace',index_label='locus_tag',index=False)
	df_infval.to_csv(pjoin('internal_data','select_mcl_results.csv'),index=False)


	## remove df_meta and df_feat dataframes from memory ##
	del [[df_infval,df_mmseqsclust]]
	gc.collect()

	## close sql connection
	conn.close()

## --------------End script---------------------##



















import os
from os.path import join as pjoin
import multiprocessing as mp
from multiprocessing import Pool
import re

## -------------- Multi-use functions---------------------##

def change_ToWorkingDirectory(directory_name):
	'''
	General function for changing directory to OUTPUT_FILES_DIR
	'''
	os.chdir(directory_name)
	print('Changing working directory to \n{}\n'.format(directory_name))


def make_OutputDirectory(new_directory):
	'''
	General function for making new output dir
	'''
	if os.path.exists(new_directory) == False:
		print('Making {} directory'.format(new_directory))
		os.mkdir(new_directory)
	else:
		print('{} directory already exists.'.format(new_directory))

def merge_ManyFiles(input_filepath,output_filepath,wildcard_search,output_filename):
	'''
	merge thousands of files. flexible for input/output directory 
	'''
	print('Merging all files with pattern {} and outputting to filename {}'.format(wildcard_search,output_filename))
	input_wildcard_search_path = pjoin(input_filepath,wildcard_search)
	output_filename_path = pjoin(output_filepath,output_filename)
	system_command = 'find {} -type f -exec cat {{}} \; > {}'.format(input_wildcard_search_path,output_filename_path)
	os.system(system_command)

def multiprocesssing_Submission(function_name,submission_list,processor_n,output_type):
	'''
	General function for multiprocessing
	'''
	pool = Pool(processes = processor_n)
	parallel_output = pool.starmap(function_name,submission_list)
	pool.close()
	if output_type == 'array':
		return(parallel_output)

def generate_AssemblyId(input_gbff_file):
	'''
	Take in a gbff filename and split it up to the second underscore. This becomes the assembly id
	'''
	assembly_id_long = re.split('_|\.',input_gbff_file)
	assembly_id = assembly_id_long[0]+'_'+assembly_id_long[1]
	return(assembly_id)


def chunks(l, n):
	'''
	Split a list into evenly sized chunks
	'''
	n = max(1, n)
	return ([l[i:i+n] for i in range(0, len(l), n)])



import os
from os.path import join as pjoin
import re
from Bio.Seq import Seq
from Bio import SeqIO, SeqFeature
from Bio.SeqRecord import SeqRecord
import pandas as pd
pd.options.mode.chained_assignment = None
import math
import numpy as np
import warnings
import time
import sqlite3 as sql
from collections import defaultdict
import gc
import sys
from gtr_utils import change_ToWorkingDirectory, make_OutputDirectory, merge_ManyFiles, multiprocesssing_Submission, generate_AssemblyId, chunks
from sklearn.cluster import DBSCAN
from sklearn.metrics.pairwise import pairwise_distances
from sklearn import metrics
import scipy.stats as stats
from scipy.cluster.hierarchy import dendrogram, linkage, cut_tree
import subprocess
from subprocess import DEVNULL, STDOUT, check_call

def cluster_DBSCAN(jac_dist, table_output_suffix, min_group_size):
	'''
	Run dbscan on a precomputed jaccard distance matrix. 
	Run through step-wise increments of epsilon values. Currently hardcoded to be 0-1 with 0.5 step increments.
	Check for case where only a single cluster is found.
	Return the 'best' clustering result. Currently as measured by max calinksi-harabasz score.
	'''
	# cluster and extract cluster label assignments

	if min_group_size == 'default':
		min_group_size = math.floor(np.log(len(jac_dist)))
	else:
		min_group_size = int(min_group_size)

	if min_group_size < 2:
		print('Minimum group size is < 2 -- will change minimum group size to 2')
		min_group_size = 2

	print('Using a minimum group size of {}'.format(min_group_size))

	cluster_label_assignments = {}
	cluster_stats=[]
	# eps_val_list=[round(n,2) for n in np.arange(0.05,1,0.05)]
	eps_val_list=[round(n,2) for n in np.arange(0.02,1,0.02)]
	for x, eps_val in enumerate(eps_val_list):
		#
		db = DBSCAN(eps=eps_val,min_samples=min_group_size,metric='precomputed').fit(jac_dist)
		core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
		core_samples_mask[db.core_sample_indices_] = True
		labels = db.labels_
		if len(set(labels))>1:
			# Number of clusters in labels, ignoring noise if present.
			n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
			n_noise_ = list(labels).count(-1)
			cluster_stats.append(
				{'eps_val':eps_val,
				'n_clusters':n_clusters_,
				'n_noise':n_noise_,
				'silh_score':round(metrics.silhouette_score(jac_dist,labels),2),
				'cali_score':round(metrics.calinski_harabasz_score(jac_dist, labels),2),
				'dbi_score': round(metrics.davies_bouldin_score(jac_dist,labels),2)})

		else:
			n_clusters_ = 1
			n_noise_ = list(labels).count(-1)
			cluster_stats.append(
				{'eps_val':eps_val,
				'n_clusters':n_clusters_,
				'n_noise':n_noise_,
				'silh_score':0,
				'cali_score':0,
				'dbi_score':0})
			# the actual per seed region cluster labels
		cluster_label_assignments['{}'.format(str(eps_val))] = pd.Series(labels)
		# Run DBSCAN again with the lowest possible input if the lowest input in the tested range has a cluster of 1
		#under construction
		print(eps_val)

	## Save two tables: the clustering statistic results and the labels for each of the clusters
	df_cluster_stats = pd.DataFrame(cluster_stats)
	df_cluster_stats.to_csv(pjoin('internal_data','dbscan_cluster_statistics{}.csv'.format(table_output_suffix)),index=False)

	df_cluster_labels = pd.DataFrame(cluster_label_assignments)
	df_cluster_labels.to_csv(pjoin('internal_data','dbscan_cluster_labels{}.csv'.format(table_output_suffix)),index=False)

	# check for case where only 1 cluster in across all tested eps values. if this is true, then return the lowest tested eps value as the 'best' value
	if df_cluster_stats['cali_score'].sum() == 0:
		best_cluster_result = df_cluster_stats.iloc[0,0]
	else:
		best_cluster_result = df_cluster_stats[df_cluster_stats['cali_score']==max(df_cluster_stats['cali_score'])]['eps_val'].iloc[0]
	df_cr = df_cluster_labels[[str(best_cluster_result)]]
	return(df_cr)



def determine_ClusterRepresentative(df_cr,df_c):
	'''
	Find the most representative seed_region for each cluster. 
	Currently using the seed_region with the lowest mean dissimilarity 
	Uses the index from df_c, which is a seed_regionX<ortho clustering id> presence/absence matrix. This makes subsequent matrix operations fast.
	IMPORTANT: Uses the jac_dist global 
	'''

	df_clusterep = pd.DataFrame()

	df_cr['seed_region_id'] = df_c.index.tolist()
	for cluster_id in sorted((df_cr.iloc[:,0]).unique()):
		# get the index postions of each seed_region_id for the current cluster
		# make a relational set of indices for the subsetted cluster
		seed_region_indices = np.array(df_cr[df_cr.iloc[:,0]==cluster_id].index.tolist())
		seed_region_indices = np.column_stack((seed_region_indices,np.asarray(range(0,len(seed_region_indices)))))

		# filter the full distance matrix for those in cluster
		interdist_array =  np.take(jac_dist,seed_region_indices[:,0],axis=0)
		interdist_array =  np.take(interdist_array,seed_region_indices[:,0],axis=1)

		n_seedregions = len(interdist_array)
		# for each row, calculate the mean dissimilarity. subtract by 1 to correct for self comparison.
		mean_dissimilarity_array = np.asarray([(sum(i)/(n_seedregions-1)) for i in interdist_array])
		mean_dissimilarity_array = np.column_stack((np.asarray(seed_region_indices),mean_dissimilarity_array))
		# sort the list to retrieve entry with lowest mean
		lowest_mean_dis_info = mean_dissimilarity_array[mean_dissimilarity_array[:,2].argsort()][0]
		# retrieve the row with dissimilarites that correspond to the region with the lowest mean dissimilarity. use the reduced index to find.
		# columns for lowest_relative_dis_info goes: full index, subsetted index, mean dissimilarity
		lowest_relative_dis = interdist_array[:,int(lowest_mean_dis_info[1])]
		## put all the pieces together into a single array:
		# add column with the representative seed_region index
		mean_dissimilarity_array = np.column_stack((np.full(shape=n_seedregions,fill_value=lowest_mean_dis_info[0]),mean_dissimilarity_array))
		# add column with cluster_id 
		mean_dissimilarity_array = np.column_stack((np.full(shape=n_seedregions,fill_value=cluster_id),mean_dissimilarity_array))
		# add column with disimilartiy values relative to the region with lowest dissimilarity
		mean_dissimilarity_array = np.column_stack((mean_dissimilarity_array,lowest_relative_dis))
		# add to master pandas dataframe. label representative uses number from full_index_pos
		colnames = ['dbscan_label','label_representative','full_index_pos','relative_index_pos','mean_dissimilarity','representative_relative_dissimilarity']
		df_clusterep = df_clusterep.append(pd.DataFrame(mean_dissimilarity_array,columns=colnames))


	# add seed_region_id to the df_clusterep dataframe. save output
	df_clusterep = df_clusterep.sort_values(by='full_index_pos')
	df_clusterep['seed_region_id'] = df_cr['seed_region_id'].tolist()
	return(df_clusterep)


def update_FinalCluster(df, cluster_dict):
	'''
	'''
	# for indices that have a clustering result, update the dictionary key.
	for c in df[df.iloc[:,0]>-1].iterrows():
		cluster_dict[c[0]] = c[1][0] # update index with cluster value
	return(cluster_dict)



def recluster_UnclusteredRegions(df_lastcluster, max_cluster_val, recluster_round):
	'''
	IMPORTANT: Uses the jac_dist global 
	'''
	# get the jaccard distance matrix index positions for all unclustered regions
	unclustered_positions = df_lastcluster[df_lastcluster.iloc[:,0] == -1]
	unclustered_positions = unclustered_positions.rename_axis('position').reset_index()
	# convert to an array that is used to filter the jaccard distance matrix
	unclustered_index = np.array(unclustered_positions.position.tolist())
	ju = jac_dist[unclustered_index[:, None], unclustered_index]
	df_unclustered = cluster_DBSCAN(jac_dist = ju, table_output_suffix = '_{}'.format(recluster_round)) 
	# add the jaccard distance matrix positions to the clustering assignments
	df_unclustered.index = unclustered_positions.position
	# update clustering assingments to be 1+ the max_cluster_val when clustering assingment is not -1
	df_unclustered.iloc[:,0] = np.where(df_unclustered.iloc[:,0] >= 0, df_unclustered.iloc[:,0] + max_cluster_val + 1, -1)
	#
	return(df_unclustered)



## --------------Start script---------------------##

def ClusterRegions(
	UserInput_main_dir,
	UserInput_output_dir_name,
	UserInput_reclustering_iterations,
	UserInput_min_group_size
	):
	'''
	'''

	# # #### TROUBLESHOOTING INPUTS #####		
	# UserInput_main_dir = '/Users/owlex/Dropbox/Documents/Northwestern/Hartmann_Lab/syntenease_project/gtr/testbed/dataset1/test1'
	# UserInput_output_dir_name = 'pdua'
	# UserInput_reclustering_iterations = 1
	# pd.set_option('display.max_columns', None)
	# # #################################		

	# make jac_dist global so it isn't copied in functions
	global jac_dist

	## read in ortholog relationship information and assign each locus_tag it's seed_region identifier##
	change_ToWorkingDirectory(directory_name = pjoin(UserInput_main_dir,UserInput_output_dir_name))
	conn = sql.connect('seed_results.db')
	df_orthos = pd.read_sql_query("SELECT * from select_mcl_results", conn)
	df_seed_regions = pd.read_sql_query("SELECT * from seed_regions", conn)

	## pseudogenes are present in the seed_regions df but not in the mcl output ##
	## assign all pseudo genes a single cluster value that is 1+ the maximum cluster_id
	df_m = df_orthos[['orf_id','ortho_cluster_id']].merge(df_seed_regions[['locus_tag','region_id']],  left_on = 'orf_id',  right_on =  'locus_tag', how = 'outer')
	df_m['orf_id'] = df_m['orf_id'].fillna(df_m['locus_tag']) #this identifies pseudo genes
	max_ortho_cluster_id_val = df_m['ortho_cluster_id'].max(skipna = True)
	df_m['ortho_cluster_id'] = df_m['ortho_cluster_id'].fillna(max_ortho_cluster_id_val+1) #this adds pseudogene identifier
	df_m['ortho_cluster_id'] = df_m['ortho_cluster_id'].astype(int)

	if len(df_m[df_m['ortho_cluster_id'].isnull()]) != 0:
		print('Some sequences dont have ortholog ids')

	## construct a jaccard distance matrix ##
	## create a presence absence table using region_id and ortho_cluster_id. cast the table so that it is binary ##
	## and then convert it to a jaccard distance ##
	df_m = df_m[['region_id','ortho_cluster_id']]
	df_m['count'] = 1
	df_m = df_m.pivot_table('count',index='region_id',columns='ortho_cluster_id').fillna(0).astype(int)
	#
	ortho_cluster_pa_array = df_m.to_numpy()
	jac_dist = pairwise_distances(ortho_cluster_pa_array, metric = "jaccard")
	## save distance matrix to contain labels and columns
	df_jd = pd.DataFrame(jac_dist,columns=df_m.index.tolist(),index=df_m.index.tolist())
	df_jd.to_csv(pjoin('internal_data','seed_regions_jaccard_dist.csv'))

	print ('Clustering {} regions'.format(len(jac_dist)))
	## Run DBSCAN on jaccard distance using different epsilon values ##
	## Returns info on clustering quality, statistics, and for the "best" cluster, dissimilarities ##
	df_cr = cluster_DBSCAN(jac_dist = jac_dist, table_output_suffix = '_0', min_group_size = UserInput_min_group_size) 


	# intialize a dictionary with all -1 clustering assignments for all cluster regions indices.
	final_cluster = {i:-1 for i in df_cr.index.tolist()}
	# add clustering results from first round
	final_cluster = update_FinalCluster(df = df_cr, cluster_dict = final_cluster)

	# reclustering procedure START
	if UserInput_reclustering_iterations > 0:
		for r in range(UserInput_reclustering_iterations):
			
			# check that -1 exist
			if sum(value == -1 for value in final_cluster.values()) <= 1:
				print('No more unclustered regions to cluster')
				break

			# Recluster
			print('Clusering round {}'.format(r + 1))
			print('Clustering {} regions'.format(sum(value == -1 for value in final_cluster.values())))
			df_cr =  recluster_UnclusteredRegions(df_lastcluster = df_cr, max_cluster_val = max(final_cluster.values()), recluster_round = r + 1)
			final_cluster = update_FinalCluster(df = df_cr, cluster_dict = final_cluster)

			# exit loop once recluster limit reached
			if r + 1 == UserInput_reclustering_iterations:
				break
	# reclustering procedure END

	# convert final_cluster dictionary to pandas df
	df_final_cluster = pd.DataFrame(final_cluster.items(), columns = ['index','clustering'])
	df_final_cluster = df_final_cluster.set_index('index')

	# find cluster representatives
	df_clusterep = determine_ClusterRepresentative(df_cr=df_final_cluster,df_c=df_m)

	## Store representative cluster labels in a new table in the seed_results sql database ##
	## also write to internal_data ##
	conn = sql.connect('seed_results.db')
	df_clusterep.to_sql(name='dbscan_label_representatives',con=conn, if_exists='replace',index_label='seed_region_id',index=False)
	df_clusterep.to_csv(pjoin('internal_data','dbscan_label_representatives.csv'),index=False)

	## remove large dataframes from memory ##
	del [[df_jd,df_clusterep,ortho_cluster_pa_array,jac_dist]]
	gc.collect()

	## close sql connection
	conn.close()


## --------------End script---------------------##





















import os
from os.path import join as pjoin
import re
import multiprocessing as mp
from multiprocessing import Pool
from Bio.Seq import Seq
from Bio import SeqIO, SeqFeature
from Bio.SeqRecord import SeqRecord
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import pandas as pd
pd.options.mode.chained_assignment = None
import numpy as np
import time
import sqlite3 as sql
from collections import defaultdict
import gc
import sys
from gtr_utils import change_ToWorkingDirectory, make_OutputDirectory, merge_ManyFiles, multiprocesssing_Submission, generate_AssemblyId, chunks



def blast_Seed(assembly_id, query_file, blast_db_path, blast_hits_path):
	'''
	'''
	full_db_path = pjoin(blast_db_path, assembly_id)
	full_hits_path = pjoin(blast_hits_path, assembly_id+'.csv')
	os.system('blastp -query {} -db {} -max_target_seqs 100 -evalue 1e-6 -outfmt "10 qseqid sseqid mismatch positive gaps ppos pident qcovs evalue bitscore qframe sframe sstart send slen qstart qend qlen" -num_threads 1 -out {}'.format(query_file, full_db_path, full_hits_path))

## write seed region gene content to file ##

def test_RegionOverlap(x1,x2,y1,y2):
	'''
	Asks: is the largest of the smallest distances less than or equal to the smallest of the largest distances?
	Returns True or False
	'''
	return( max(x1,y1) <= min(x2,y2) )



def filter_BlastHits(df, identity_cutoff, coverage_cutoff):
	'''
	Remove hits that do not meet identity and coverage cutoffs
	'''
	df = df[ (df['pident'] >= identity_cutoff ) & (df['qcovs'] >= coverage_cutoff) ]
	return(df)



def extract_SeedRegions(assembly_id, upstream_search_length, downstream_search_length, identity_cutoff, coverage_cutoff, hits_cutoff):
	'''
	Check for overlaps in seed hits. When this occurs, pick the best hit for overlap.
	Extract x basepairs upstream and downstream of a seed blast hit.
	Write to sql database under table 'seed_regions'
	'''

#--------- Troubleshooting --------#
# pd.set_option('display.max_columns', None)
# upstream_search_length, downstream_search_length, identity_cutoff, coverage_cutoff, hits_cutoff = 10000,10000,20,80, 1

# # UserInput_main_dir = '/projects/b1042/HartmannLab/alex/GeneGrouper_test/testbed/dataset1/test1'#'/Users/owlex/Dropbox/Documents/Northwestern/Hartmann_Lab/syntenease_project/gtr/testbed/dataset1/test1'
# UserInput_main_dir = '/Users/owlex/Dropbox/Documents/Northwestern/Hartmann_Lab/syntenease_project/gtr/testbed/dataset3/test1'
# UserInput_output_dir_name = pjoin(UserInput_main_dir,'pdua')
# os.chdir(UserInput_main_dir)
# conn = sql.connect(pjoin(UserInput_main_dir,'genomes.db')) # genome database holds all base info
# conn2 = sql.connect(pjoin(UserInput_output_dir_name,'seed_results.db')) # seed_results database holds all seed search specific info. 


# assembly_id = 'GCF_009800085'  #GCF_009800085_02009     dbscan_label 3     ,global_strand -1 has more
# assembly_id = 'GCF_000583895' #GCF_000583895_02788     dbscan_label 13    ,global_strand 1  has fewer

# assembly_id = 'GCF_001077835'

# assembly_id = 'GCF_000251025'

#--------- Troubleshooting --------#

	try:
		df_hits = pd.read_csv(pjoin('temp_blast_results',assembly_id+'.csv'),names=['qseqid','sseqid','mismatch', 'positive','gaps', 'ppos','pident','qcovs','evalue','bitscore','qframe','sframe','sstart', 'send', 'slen', 'qstart', 'qend', 'qlen'])
	except:
		return(pd.DataFrame())
		pass

	## filter hit results, exit function if not hits pass cutoffs
	df_hits = filter_BlastHits(df=df_hits, identity_cutoff=identity_cutoff, coverage_cutoff=coverage_cutoff)
	if len(df_hits) == 0:
		return(pd.DataFrame())
		pass


	## read in whole genome info
	df_g = pd.read_sql_query("SELECT contig_id, locus_tag, cds_start, cds_end, strand, pseudo_check from gb_features WHERE assembly_id = '{}' ".format(assembly_id), sql.connect('genomes.db'))


	## merge genome level data with blast hit results. sort by identity and evalue so that best hits are at the top of the table
	df_hits = df_g.merge(df_hits[['sseqid','pident','qcovs','evalue']],left_on='locus_tag',right_on='sseqid',how='inner')


	# if there are multiple hits that are identical, keep the one with the highest evalue
	df_hits = df_hits.groupby('locus_tag').first().reset_index()

	df_hits = df_hits.sort_values(['pident','evalue'],ascending=[False,False])

	# keep only n total hits (default is all hits)
	if type(hits_cutoff) == int:
		df_hits = df_hits.iloc[:hits_cutoff, :]


	## define upstream and downstream search lengths. The distance inputs for upstream/start and downstream/end are switched when the seed gene is in the -1 frame
	## The usl and dsl are not strand specific now. They are relative to the genomes' upstream and downstream positions
	df_hits['usl'] = np.where(df_hits['strand']==1, df_hits['cds_start'] - upstream_search_length, df_hits['cds_start'] - downstream_search_length)
	df_hits['dsl'] = np.where(df_hits['strand']==1, df_hits['cds_end'] + downstream_search_length, df_hits['cds_end'] + upstream_search_length)

	## for each contig, test for overlapping region ranges. Keep the region with the best hit.
	## append subsets to new df called df_hits_parsed
	df_hits_parsed = pd.DataFrame()
	for cid in df_hits['contig_id'].unique().tolist():
		#Looping over contigs to prevent instances where they overlap but are not on the same contig.
		df_hits_contig = df_hits[df_hits['contig_id']==cid]
		for hid in df_hits_contig['locus_tag']:
			try:
				# create range to compare all other ranges to			
				hid_x1, hid_x2 = df_hits_contig[df_hits_contig['contig_id']==cid][df_hits_contig['locus_tag']==hid]['usl'].item(), df_hits_contig[df_hits_contig['contig_id']==cid][df_hits_contig['locus_tag']==hid]['dsl'].item()
				# test for overlap with all other ranges on the same contig
				df_hits_contig['overlap'] = df_hits_contig[df_hits_contig['contig_id']==cid].apply(lambda x: test_RegionOverlap(x1=hid_x1, x2=hid_x2, y1=x['usl'], y2=x['dsl']), axis=1)
				keep_index  = (df_hits_contig['overlap']==True).idxmax()
				df_hits_contig['overlap_representative'] = np.where(df_hits_contig.index==keep_index,True,False)
				# remove rows that are not repesentatives but do overlap. keep everything else.
				df_hits_contig = df_hits_contig[ (df_hits_contig['overlap']==False) | (df_hits_contig['overlap_representative'] == True) | (df_hits_contig['overlap']).isnull() == True ]
			except:
				continue
		df_hits_parsed = df_hits_parsed.append(df_hits_contig)



	## build a dataframe that contains genes upsteam and downstream of boundaries for each seed blast hit ##
	df_rkeep = pd.DataFrame()
	for h in df_hits_parsed.iterrows():


		# get coordinates for upstream and downstream gene boundaries

		# ----- attempt start ----- #
		# cds_start for strand == 1 
		# cds_end for strand == -1
		if h[1][4] == 1:
			cds_search_list = df_g[df_g['contig_id'] == h[1][1]]['cds_start'].tolist()
			usl_locus_tag_bp = min(cds_search_list, key = lambda x: abs(x-h[1][10]))

			cds_search_list = df_g[df_g['contig_id'] == h[1][1]]['cds_end'].tolist()
			dsl_locus_tag_bp = min(cds_search_list, key = lambda x: abs(x-h[1][11]))

		if h[1][4] == -1:
			cds_search_list = df_g[df_g['contig_id'] == h[1][1]]['cds_start'].tolist()
			usl_locus_tag_bp = min(cds_search_list, key = lambda x: abs(x-h[1][10]))

			cds_search_list = df_g[df_g['contig_id'] == h[1][1]]['cds_end'].tolist()
			dsl_locus_tag_bp = min(cds_search_list, key = lambda x: abs(x-h[1][11]))



		# subset the main genome dataframe to contain only ORFs that are within the designated bounds
		if h[1][4] == 1:	
			df_r = df_g[ df_g['contig_id'] == h[1][1] ][ ( df_g['cds_end'] >= usl_locus_tag_bp) & (df_g['cds_start'] <= dsl_locus_tag_bp ) ]
		if h[1][4] == -1:	
			df_r = df_g[ df_g['contig_id'] == h[1][1] ][ ( df_g['cds_end'] >= usl_locus_tag_bp) & (df_g['cds_start'] <= dsl_locus_tag_bp ) ]
		# ----- attempt end ----- #

		# append the subsetted dataframe to a new dataframe that will contain all region extractions
		df_r['region_id'] = h[1][0]
		df_rkeep = df_rkeep.append(df_r,ignore_index=True)

	## add blast data to the extracted regions, remove unneeded columns, and return the dataframe
	df_rkeep['assembly_id'] = assembly_id
	df_rkeep = df_rkeep[['region_id','assembly_id','contig_id','locus_tag','strand','pseudo_check']]
	df_rkeep = df_rkeep.merge(df_hits_parsed[['sseqid','pident','qcovs','evalue']], left_on='region_id',right_on='sseqid')
	df_rkeep = df_rkeep.drop(columns='sseqid')


	return(df_rkeep)


def write_RegionSeqsToFile(assembly_id, output_dir_name):
	'''
	Write a .faa containing all sequences that were found to be within a defined seed region for a given assembly_id
	
	'''
	df_s =  pd.read_sql_query("SELECT locus_tag from seed_regions WHERE assembly_id = '{}'".format(assembly_id), conn2)
	if len(df_s) == 0:
		return()

	seqs_to_write = df_s['locus_tag'].tolist()
	with open(pjoin(output_dir_name,'regions',assembly_id+'.faa'),'w') as outfile:
		for record in SeqIO.parse(pjoin('assemblies',assembly_id+'.faa'),'fasta'):
			if record.id in seqs_to_write:
				outfile.write('>{}\n'.format(record.id))
				outfile.write('{}\n'.format(str(record.seq)))




def BuildQueryFiles(
	UserInput_main_dir,
	UserInput_genome_inputs_dir,
	UserInput_output_dir_name,
	UserInput_query_filename_path,
	UserInput_upstream_search_length,
	UserInput_downstream_search_length,
	UserInput_identity_threshold,
	UserInput_coverage_threshold,
	UserInput_hitcount_threshold,
	UserInput_processes,
	):
	'''
	'''

#--------- Troubleshooting --------#
# pd.set_option('display.max_columns', None)
# mp.set_start_method("fork")
# # UserInput_main_dir = '/projects/b1042/HartmannLab/alex/GeneGrouper_test/testbed/dataset1/test1'#'/Users/owlex/Dropbox/Documents/Northwestern/Hartmann_Lab/syntenease_project/gtr/testbed/dataset1/test1'
# UserInput_main_dir = '/Users/owlex/Dropbox/Documents/Northwestern/Hartmann_Lab/syntenease_project/gtr/testbed/dataset3/test1'
# UserInput_output_dir_name = pjoin(UserInput_main_dir,'pdua')
# UserInput_genome_inputs_dir = '/Users/owlex/Dropbox/Documents/Northwestern/Hartmann_Lab/syntenease_project/gtr/testbed/dataset3/core'
# UserInput_query_filename_path = pjoin(UserInput_genome_inputs_dir,'pdua.txt')
# UserInput_upstream_search_length = 2000
# UserInput_downstream_search_length = 18000
# UserInput_identity_threshold = 15
# UserInput_coverage_threshold = 70
# UserInput_hitcount_threshold = 5
# UserInput_processes = 8
# os.chdir(UserInput_main_dir)
# # conn = sql.connect(pjoin(UserInput_main_dir,'genomes.db')) # genome database holds all base info
# # conn2 = sql.connect(pjoin(UserInput_output_dir_name,'seed_results.db')) # seed_results database holds all seed search specific info. 
#--------- Troubleshooting --------#


	make_OutputDirectory(new_directory=UserInput_main_dir)
	change_ToWorkingDirectory(directory_name=UserInput_main_dir)
	make_OutputDirectory(new_directory=UserInput_output_dir_name)
	make_OutputDirectory(new_directory=pjoin(UserInput_output_dir_name,'regions'))
	make_OutputDirectory(new_directory=pjoin(UserInput_output_dir_name,'ortholog_clusters'))
	make_OutputDirectory(new_directory=pjoin(UserInput_output_dir_name,'internal_data'))
	make_OutputDirectory(new_directory=pjoin(UserInput_output_dir_name,'subgroups'))
	make_OutputDirectory(new_directory=pjoin(UserInput_output_dir_name,'visualizations'))
	make_OutputDirectory(new_directory='temp_blast_results')

	global conn, conn2

	conn = sql.connect(pjoin(UserInput_main_dir,'genomes.db')) # genome database holds all base info
	conn2 = sql.connect(pjoin(UserInput_output_dir_name,'seed_results.db')) # seed_results database holds all seed search specific info. 


	##---- Blast seed against database ----- ##
	r_params = [[generate_AssemblyId(input_gbff_file=f), UserInput_query_filename_path, 'blast_database', 'temp_blast_results'] for f in os.listdir(UserInput_genome_inputs_dir) if f.endswith('.gbff')]
	print('Blasting seed against {} genomes'.format(len(r_params)))
	start_t = time.time()
	with Pool(processes=UserInput_processes) as p:
		p.starmap(blast_Seed, r_params[:])
	print((time.time()-start_t)/60)


	##---- identify and extract seed regions ----- ##
	r_params = [[generate_AssemblyId(input_gbff_file=f), UserInput_upstream_search_length, UserInput_downstream_search_length,UserInput_identity_threshold,UserInput_coverage_threshold, UserInput_hitcount_threshold] for f in os.listdir(UserInput_genome_inputs_dir) if f.endswith('.gbff')]
	print('Identifying and extracting seed regions for {} genomes'.format(len(r_params)))
	start_t = time.time()
	df_regions = pd.DataFrame()
	with Pool(processes=UserInput_processes) as p:
		df_regions = pd.concat(p.starmap(extract_SeedRegions, r_params[:]))
	print((time.time()-start_t)/60)
	print('mem usage:', df_regions.memory_usage(deep=True).sum()/1028/1028)

	if len(df_regions) == 0:
		raise ValueError('No genes matched your query!')
		
	df_regions.to_sql(name='seed_regions',con=conn2, if_exists='replace',index_label='locus_tag',index=False)


	##---- index seed_regions from seed_results databse on assembly_id to speed up searches ----- ##
	c2 = conn2.cursor()
	c2.execute('CREATE INDEX assembly_id_index on seed_regions(assembly_id)') # this is key to speeding up search
	conn2.commit()


	##---- write seed region gene content to sql database ----- ##
	r_params = [[generate_AssemblyId(input_gbff_file=f), UserInput_output_dir_name] for f in os.listdir(UserInput_genome_inputs_dir) if f.endswith('.gbff')]
	print('Writing seed region sequences for for {} genomes'.format(len(r_params)))
	start_t = time.time()
	df_regions = pd.DataFrame()
	with Pool(processes=UserInput_processes) as p:
		p.starmap(write_RegionSeqsToFile, r_params[:])
	print((time.time()-start_t)/60)

	## merge sequence files and write to ortholog_clusters df
	merge_ManyFiles(input_filepath=pjoin(UserInput_output_dir_name,'regions'),output_filepath=pjoin(UserInput_output_dir_name,'ortholog_clusters'),wildcard_search='*.faa',output_filename='merged_aa.faa')

	## delete ./temp_blast_results and <seed_region>/regions
	os.system('rm -R ./temp_blast_results')
	#### os.system('rm -R ./{}/regions'.format(UserInput_output_dir_name))


	conn.close()
	conn2.close()

	del conn
	del conn2


## ---- END OF SCRIPT ---- ##




	
import os
from os.path import join as pjoin
import re
import multiprocessing as mp
from multiprocessing import Pool
from Bio.Seq import Seq
from Bio import SeqIO, SeqFeature
from Bio.SeqRecord import SeqRecord
import pandas as pd
pd.options.mode.chained_assignment = None
import numpy as np
import warnings
import time
import sqlite3 as sql
from collections import defaultdict
import gc
from gtr_utils import change_ToWorkingDirectory, make_OutputDirectory, merge_ManyFiles, multiprocesssing_Submission, generate_AssemblyId, chunks
import subprocess
from subprocess import DEVNULL, STDOUT, check_call

def extract_GenbankMetadata(input_filename, gbff_raw_dir):
	'''
	Extracts the organism name from the genbank file organism annotation. 
	If there is only a one line name under the organism then 'Undefined' gets added after the name as a filler for species.
	'''
	assembly_id = generate_AssemblyId(input_gbff_file=input_filename)
	for x, record in enumerate(SeqIO.parse(pjoin(gbff_raw_dir,input_filename),'gb')):
		if x == 0:
			df_tax = pd.DataFrame({'assembly_id':[assembly_id], 'filename':[input_filename]})
			try:
				df_tax['organism'] = record.annotations['organism'] + ' Undefined'
				# checking name 
				# organism_name = record.annotations['organism']
				# name_list = organism_name.split(' ')
				# if len(count_whitespace) == 1:
				# 	organism_name = organism_name + ' ' + 'Undefined'
				#  df_tax['organism'] = organism_name
			except:
				df_tax['organism'] = 'UndefinedGenus UndefinedSpecies'
		break	
	return(df_tax)



def extract_GenbankFeatures(input_filename, gbff_raw_dir, assembly_faa_dir):
	'''
	Large function to process  that extracts all relevant genbank data
	Extracts CDS features
	Writes all CDS to ./assembly folder
	'''
	assembly_id = generate_AssemblyId(input_gbff_file=input_filename)
	store_rows = []
	with open(pjoin(assembly_faa_dir, assembly_id+'.faa'),'w') as outfile:
		n = 0
		for x, record in enumerate(SeqIO.parse(pjoin(gbff_raw_dir,input_filename),'gb')):
			## extract releavant CDS features ##
			for f in record.features:
				if f.type=='CDS':
					try:
						refseq_locus_tag = f.qualifiers['locus_tag'][0]
					except:
						refseq_locus_tag = 'NA'
					try:
						refseq_product = f.qualifiers['product'][0]
					except:
						refseq_product = 'NA'
					try:
						refseq_gene = f.qualifiers['gene'][0]
					except:
						refseq_gene = 'NA'
					try: 
						reseq_translation = f.qualifiers['translation'][0]
					except:
						reseq_translation = 'NA'
					##
					if reseq_translation == 'NA':
						pseudo_check = 'p'
					else:
						pseudo_check = 'g'
					#internal assembly id, internal contig id, internal locus tag, refseq locus tag, cds start, cds end, cds strand, refseq product, refseq gene
					locus_tag_string = str(n).zfill(5) # generates a buffer of up to 5 zeroes so all locus_tag have same length
					locus_tag = assembly_id+'_'+locus_tag_string
					# check that f.location gives the cds coordinates and not the whole contig coordinates
					if len(f.location.parts) > 1:
						start_cds = f.location.parts[1].start
						end_cds = f.location.parts[1].end
					else:
						start_cds = f.location.start.position
						end_cds = f.location.end.position
					# store in list
					store_rows.append(
						[
						assembly_id,
						assembly_id+'_'+str(x), #contig_id
						locus_tag,
						refseq_locus_tag,
						start_cds,
						end_cds,
						f.location.strand,
						refseq_product,
						refseq_gene,
						pseudo_check
						])
					n+=1
					if reseq_translation == 'NA':
						continue
					## Write CDS to file ##
					outfile.write('>{} {}\n'.format(locus_tag, refseq_locus_tag))
					outfile.write('{}\n'.format(reseq_translation))
		## return genbank features
		df_gbf = pd.DataFrame(store_rows,columns=['assembly_id','contig_id','locus_tag','refseq_locus_tag','cds_start','cds_end','strand','refseq_product','refseq_gene','pseudo_check'])
		return(df_gbf)


def make_BlastDatabase(assembly_id, assembly_faa_dir, blast_db_path):
	'''
	'''
	full_faa_path = pjoin(assembly_faa_dir,assembly_id+'.faa')
	full_db_path = pjoin(blast_db_path,assembly_id)
	check_call(['makeblastdb', '-in', full_faa_path, '-out', full_db_path, '-dbtype', 'prot', '-title', '"{}_db"'.format(assembly_id), '-parse_seqids'], stdout=DEVNULL, stderr=STDOUT)



def update_GenomesDatabase():
	'''
	'''
	pass


def check_ForGenomes(genome_inputs_dir):
	'''
	'''

	pass


def build_GenomesDatabase(genomes_list):
	'''
	'''
	pass






## ---- START OF SCRIPT ---- ##
def BuildGenomeDatabase(UserInput_main_dir, UserInput_genome_inputs_dir, UserInput_processes):
	'''
	'''

	# make directory, with subdirectories to store data. and database for genomic data
	make_OutputDirectory(new_directory=UserInput_main_dir)
	change_ToWorkingDirectory(directory_name=UserInput_main_dir)
	make_OutputDirectory(new_directory='blast_database')
	make_OutputDirectory(new_directory='assemblies')
	conn = sql.connect(pjoin(UserInput_main_dir,'genomes.db')) # genome database holds all base info


	##---- Extract organism info ----- ##
	if os.path.isfile('genomes.db'):
		try:
			# if df_meta exists, then check for which files to be processed are not present in df_meta, if any
			df_meta = pd.read_sql_query("SELECT assembly_id from gb_metadata",conn)
			meta_exists_list = df_meta.assembly_id.tolist()
			r_params = [[f, UserInput_genome_inputs_dir] for f in os.listdir(UserInput_genome_inputs_dir) if f.endswith('.gbff') if generate_AssemblyId(input_gbff_file=f) not in meta_exists_list]
		except:
			# if there is an error reading in df_meta, then it doesn't exist. Therefore, process all input files 
			r_params = [[f, UserInput_genome_inputs_dir] for f in os.listdir(UserInput_genome_inputs_dir) if f.endswith('.gbff')]



	if len(r_params) != 0:
		chunk_r_params_input = chunks(l=r_params, n=10)
		print('Extracting metadata for {} genomes'.format(len(r_params)))
		start_t = time.time()
		df_meta = pd.DataFrame()
		with Pool(processes=UserInput_processes) as p:
			df_meta = pd.concat(p.starmap(extract_GenbankMetadata, r_params[:]))
		df_meta.to_sql(name='gb_metadata',con=conn, if_exists='append',index_label='assembly_id',index=False)
		print((time.time()-start_t)/60)

	else:
		print('All files already have metadata extracted')



	##---- Extract gene features ----- ##
	if os.path.isfile('genomes.db'):
		try:
			# if features have already been pre-processed, then check which files, if any, need to be processed
			df_feat = pd.read_sql_query("SELECT DISTINCT assembly_id from gb_features",conn)
			feat_exist_list = df_feat.assembly_id.tolist()
			r_params = [[f, UserInput_genome_inputs_dir, 'assemblies'] for f in os.listdir(UserInput_genome_inputs_dir) if f.endswith('.gbff') if generate_AssemblyId(input_gbff_file=f) not in meta_exists_list]
		except:
			# if there is an error read in df_feat, then it doesn't exist and all files need to be processed
			r_params = [[f, UserInput_genome_inputs_dir, 'assemblies'] for f in os.listdir(UserInput_genome_inputs_dir) if f.endswith('.gbff')]


	update_sql_index = False
	if len(r_params) != 0:

		print('Extracting features for {} genomes'.format(len(r_params)))
		start_t = time.time()
		chunk_r_params_input = chunks(l=r_params, n=100) # Use a chunking strategy to reduce the memory consumption caused by having large pandas dataframes
		for chunk_r in chunk_r_params_input:
			df_feat = pd.DataFrame()
			with Pool(processes=UserInput_processes) as p:
				df_feat = pd.concat(p.starmap(extract_GenbankFeatures, chunk_r)) 
			df_feat.to_sql(name='gb_features',con=conn, if_exists='append',index_label='locus_tag',index=False)
			print('mem usage:', df_feat.memory_usage(deep=True).sum()/1028/1028)
		print((time.time()-start_t)/60)
		update_sql_index = True

	else:
		print('All files already have features extracted')


	## remove df_meta and df_feat dataframes from memory ##
	del [[df_meta,df_feat]]
	gc.collect()
	df_meta=pd.DataFrame()
	df_feat=pd.DataFrame()


	## index gb_features from genome database on assembly_id to speed up searches ##
	if update_sql_index == True:
		try:
			c = conn.cursor()
			c.execute('CREATE INDEX assembly_id_index on gb_features(assembly_id)')
			conn.commit()
			print('Making assembly_id into index for gb_features')
		except:
			print('Did not update gb_features index')
			pass


	##---- Write blast database files ----- ##
	#
	#check which genomes already have blast databases constructed by comparing blast file to files in df_feat
	df_feat = pd.read_sql_query("SELECT DISTINCT assembly_id from gb_features",conn) 
	feat_exist_list = df_feat.assembly_id.tolist()
	blastdb_files = [x[:-4] for x in os.listdir('blast_database') if x.endswith('.pog')]
	files_to_blastdb = [b for b in feat_exist_list if b not in blastdb_files ]


	if len(files_to_blastdb) != 0:

		print('Making blast database for {} genomes'.format(len(r_params)))
		r_params = [[b,'assemblies','blast_database'] for b in files_to_blastdb]
		start_t = time.time()
		with Pool(processes=UserInput_processes) as p:
			p.starmap(make_BlastDatabase, r_params[:])
		print((time.time()-start_t)/60)
	else:
		print('All files already have blast databases constructed')


	conn.close()

	del conn

## ---- END OF SCRIPT ---- ##





#!/usr/bin/env python

import warnings
warnings.simplefilter(action='ignore', category=UserWarning)
	
import os
from os.path import join as pjoin
import time
import argparse
import sys
import multiprocessing as mp
from multiprocessing import Pool

mp.set_start_method("fork")
global script_path

script_path = os.path.split(os.path.realpath(os.path.abspath(__file__)))[0]
sys.path.append(script_path)

from gtr_utils import change_ToWorkingDirectory, make_OutputDirectory, merge_ManyFiles, multiprocesssing_Submission, generate_AssemblyId, chunks
from build_GenomeDatabase import *
from build_QueryFiles import *
from cluster_RegionSeqs import *
from cluster_Regions import *
from output_RegionResults import *
from make_Visualizations import *


def run_BuildGenomeDatabase(args):
	'''
	'''
	print('run_BuildGenomeDatabase with {} threads\n'.format(args.threads))
	MainDir = os.path.abspath(args.project_directory)
	GenomeInputsDir = os.path.abspath(args.genomes_directory)

	if os.path.isdir(MainDir) == False:
		try:
			os.mkdir(MainDir)
		except:
			raise ValueError('main directory not found')
	if os.path.isdir(GenomeInputsDir) == False:
		raise ValueError('genomes directory not found')

	start_t = time.time()
	BuildGenomeDatabase(
	UserInput_genome_inputs_dir=GenomeInputsDir, 
	UserInput_main_dir=MainDir, 
	UserInput_processes=args.threads
	)
	print('\nFinished building database\nTotal time (m): {}'.format((time.time()-start_t)/60))

def run_RegionSearch(args):
	'''
	'''
	print('run_RegionSearch with {} threads\n'.format(args.threads))


	MainDir = os.path.abspath(args.project_directory)
	GenomeInputsDir = os.path.abspath(args.genomes_directory)
	QueryFilenamePath = os.path.abspath(args.query_file)

	if os.path.isdir(MainDir) == False:
		raise ValueError('main directory not found')
	if os.path.isdir(GenomeInputsDir) == False:
		raise ValueError('genomes directory not found')
	if os.path.isfile(QueryFilenamePath) == False:
		raise ValueError('seed file not found')
	if os.path.isdir(pjoin(MainDir,args.search_name)) == True:
		if args.force == True:
			os.system('rm -R {}'.format(pjoin(MainDir,args.search_name)))
		else:
			raise ValueError('search directory already exists')
			print('error!!!')

	start_t = time.time()
	BuildQueryFiles(
		UserInput_main_dir = MainDir,#args.project_directory, 
		UserInput_genome_inputs_dir= GenomeInputsDir,#args.genomes_directory,
		UserInput_output_dir_name = args.search_name,
		UserInput_query_filename_path = QueryFilenamePath,
		UserInput_upstream_search_length = args.upstream_search,
		UserInput_downstream_search_length = args.downstream_search,
		UserInput_identity_threshold = args.seed_identity,
		UserInput_coverage_threshold = args.seed_coverage,
		UserInput_hitcount_threshold = args.seed_hits_kept,
		UserInput_processes = args.threads
		)

	ClusterRegionSeqs(
		UserInput_main_dir = MainDir,#args.project_directory, 
		UserInput_output_dir_name = args.search_name,
		UserInput_faa_chunk_size = 10,
		UserInput_blast_filter_params = [50,95], #identity, qcoverage
		UserInput_inflation_values = ['1.2','1.4','2.0','4.0','6.0'],
		UserInput_similarity_scoring_method = 'evalue',
		UserInput_linclust_settings = '--min-seq-id 0.95',
		UserInput_ortho_select_method = 'max_div',
		UserInput_processes = args.threads
		)

	ClusterRegions(
		UserInput_main_dir = MainDir,#args.project_directory, 
		UserInput_output_dir_name = args.search_name,
		UserInput_reclustering_iterations = args.recluster_iterations,
		UserInput_min_group_size = args.min_group_size
		)

	OutputRegionResults(
		UserInput_main_dir = MainDir,#args.project_directory, 
		UserInput_output_dir_name = args.search_name,
		)

	print('\nFinished searching for gene clusters\nTotal time (m): {}'.format((time.time()-start_t)/60))


def run_MakeVisualizations(args):
	print('make_Visualizations\n')


	MainDir = os.path.abspath(args.project_directory)
	if os.path.isdir(MainDir) == False:
		raise ValueError('main directory not found')

	if args.visual_type == 'main':
		print('Generating main visualization')
	if args.visual_type == 'region_cluster':
		print('Generating region cluster visualization for cluster label {}'.format(args.cluster_label))

	start_t = time.time()
	MakeVisualizations(
		UserInput_main_dir = MainDir,
		UserInput_output_dir_name = args.search_name,
		UserInput_cluster_label_id = args.group_label,
		UserInput_visualization_type = args.visual_type,
		UserInput_script_path = pjoin(script_path,'Rscripts'),
		UserInput_image_format = args.image_format,
		UserInput_tip_label_type = args.tip_label_type,
		UserInput_tip_label_size = args.tip_label_size
		)

	print('\nFinished making visualization\nTotal time (m): {}'.format((time.time()-start_t)/60))



def main(args = None):
	'''
	'''
	if args is None:
		args = sys.argv[1:]

	# create parser
	parser = argparse.ArgumentParser(prog = 'GeneGrouper')

	# add common args to parser
	parser.add_argument('-d', '--project_directory', type=str, default=os.getcwd(), help = 'Main directory to contain the base files used for region searching and clustering. Default current directory.', metavar='')
	parser.add_argument('-n', '--search_name', type=str, default = 'region_search', help = 'Name of the directory to contain search-specific results. Default region_search', metavar='')
	parser.add_argument('-g', '--genomes_directory', type=str, default=pjoin(os.getcwd(),'genomes'), help = 'Directory containing genbank-file format genomes with the suffix .gbff. Default ./genomes.', metavar='')
	parser.add_argument('-t', '--threads', type=int, default= mp.cpu_count(), help = 'Number of threads to use. Default all threads.', metavar='')

	# make subparser
	subparsers = parser.add_subparsers(title='subcommands', description ='Valid subcommands', help = 'sub-command help', dest='command')

	# build_database subparser 
	parser_bd = subparsers.add_parser('build_database', help ='Convert a set of genomes into a useable format for GeneGrouper')
	parser_bd.set_defaults(func=run_BuildGenomeDatabase)

	# find_regions args
	parser_fr = subparsers.add_parser('find_regions', help ='Find regions given a translated gene and a set of genomes')
	parser_fr.add_argument('-f', '--query_file', type=str, required = True, help = 'Provide the absolute path to a fasta file containing a translated gene sequence.', metavar='')
	parser_fr.add_argument('-us', '--upstream_search', type=int, default=10000, help = 'Upstream search length in basepairs. Default=10000', metavar='')
	parser_fr.add_argument('-ds', '--downstream_search', type=int, default=10000, help = 'Downstream search length in basepairs. Default=10000', metavar='')
	parser_fr.add_argument('-i', '--seed_identity', type=int, default=60, help = 'Identity cutoff for initial blast search. Default=60', metavar='')
	parser_fr.add_argument('-c', '--seed_coverage', type=int, default=90, help = 'Coverage cutoff for initial blast search. Default=90', metavar='')
	parser_fr.add_argument('-hk', '--seed_hits_kept', type = int, default = None, help = 'Number of blast hits to keep. Default=None', metavar='')
	parser_fr.add_argument('--min_group_size', default = 'default', help = 'The minimum number of gene regions to constitute a group. Default=ln(jaccard distance length)', metavar='')
	parser_fr.add_argument('-re', '--recluster_iterations', type = int, default = 0, help = 'Number of iterations to run DBSCAN on remaining ungrouped regions . Default=0', metavar='')
	parser_fr.add_argument('--force', action='store_true', help = 'Flag to overwrite search name directory.')
	parser_fr.set_defaults(func=run_RegionSearch)

	# visualize_regions args
	parser_mv = subparsers.add_parser('visualize', help ='Visualize GeneGrouper outputs. Three visualization options are provided. Check the --visual_type help description.')
	parser_mv.add_argument('--visual_type', type=str, choices = ['main','group','tree'], default = 'main', help="Choices: [main, group, tree]. Use main for main visualizations. Use group to inspect specific group. Use tree for a phylogenetic tree of representative seed sequencess. Default=main", metavar='')
	parser_mv.add_argument('--group_label', type=int, default=-1, help='The integer identifier of the group you wish to inspect. Default=-1',metavar='')
	parser_mv.add_argument('--image_format', type=str, choices = ['png','svg'], default='png', help='Choices: [png,svg]. Output image format. Use svg if you want to edit the images. Default=png.',metavar='')
	parser_mv.add_argument('--tip_label_type', type=str, choices = ['full','group'], default='full', help='Choices: [full, group]. Use full to include the sequence ID followed by group ID. Use group to only have the group ID. Default=full',metavar='')
	parser_mv.add_argument('--tip_label_size', type=int, default=2, help='Specify the tip label size in the output image. Default=2',metavar='')

	parser_mv.set_defaults(func=run_MakeVisualizations)

	args = parser.parse_args()
	args.func(args)


## -------------- Run ---------------------##

if __name__ == '__main__':
	sys.exit(main())

#os.system('conda list --export > list.txt')





import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
	long_description = fh.read()

setuptools.setup(
	name="GeneGrouper",
	version="1.0.3",
	author="Alexander G. McFarland",
	author_email="alexandermcfarland2022@u.northwestern.edu",
	description="Find and cluster genomic regions containing a seed gene",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/agmcfarland/GeneGrouper",
	project_urls={
		"Bug Tracker": "https://github.com/agmcfarland/GeneGrouper/issues",
	},
	classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
	],
	package_dir={"": "src"},
	packages=setuptools.find_packages(where="src", exclude=['docs','test_data']), #exlude = ['*.egg-info', ]
	entry_points={
		'console_scripts': ['GeneGrouper = GeneGrouper.__main__:main'], },
	package_data={'GeneGrouper' : ['Rscripts/*']},
	python_requires=">=3.6")

# https://packaging.python.org/tutorials/installing-packages/#creating-virtual-environments
# https://packaging.python.org/tutorials/packaging-projects/
# https://trstringer.com/easy-and-nice-python-cli/
# https://github.com/pypa/sampleproject/blob/main/setup.py
# https://chriswarrick.com/blog/2014/09/15/python-apps-the-right-way-entry_points-and-scripts/




import os
from os.path import join as pjoin
import re
from Bio.Seq import Seq
from Bio import SeqIO, SeqFeature
from Bio.SeqRecord import SeqRecord
import pandas as pd
pd.options.mode.chained_assignment = None
import numpy as np
import warnings
import time
import sqlite3 as sql
from collections import defaultdict
import gc
import sys
from gtr_utils import change_ToWorkingDirectory, make_OutputDirectory, merge_ManyFiles, multiprocesssing_Submission, generate_AssemblyId, chunks
from output_RegionResults import subset_SimilarityMatrix
import subprocess
from subprocess import DEVNULL, STDOUT, check_call

def convert_ToGGGenesInput(df_m,filter_on):
	'''
	'''
	df_gggenes = pd.DataFrame()
	for clabel in df_m[filter_on].unique().tolist():
		print(str(int(clabel)))
		df_cl = df_m[df_m[filter_on] == clabel]
		df_cl = df_cl.sort_values(by=['new_pos'],ascending=True)

		# align genes to normalized position of the seed
		start_bp = df_cl[df_cl['is_seed']==1]['cds_start1'].item()

		# process downstream list
		cds_list = list()
		for row in df_cl[df_cl['new_pos']>=0].iterrows():
			row_pos1 = len(row[1])-2
			row_pos2 = len(row[1])-1
			cds_list.append(row[1][row_pos1])
			cds_list.append(row[1][row_pos2])

		downstream_xlist = [0]
		while len(cds_list) >1:
			if cds_list[0] == start_bp:
				x1 = cds_list[1] - cds_list[0]
				downstream_xlist.append(x1)
				del cds_list[0]
				continue

			x1 = cds_list[1] - cds_list[0] + x1
			del cds_list[0]
			downstream_xlist.append(x1)

		# process upstream list
		cds_list = list()
		for row in df_cl[df_cl['new_pos']<=0].iterrows():
			row_pos1 = len(row[1])-2
			row_pos2 = len(row[1])-1
			cds_list.append(row[1][row_pos1])
			cds_list.append(row[1][row_pos2])

		cds_list.reverse()
		del cds_list[0]
			
		upstream_xlist = [0]
		while len(cds_list) >1:
			if cds_list[0] == start_bp:
				x1 = cds_list[1] - cds_list[0]
				upstream_xlist.append(x1)
				del cds_list[0]
				continue

			x1 = cds_list[1] - cds_list[0] + x1
			del cds_list[0]
			upstream_xlist.append(x1)
		del upstream_xlist[0]
		upstream_xlist.reverse()

		# map values back to dataframe
		upstream_xlist = np.asarray(upstream_xlist).reshape(int(len(upstream_xlist)/2),2)
		downstream_xlist = np.asarray(downstream_xlist).reshape(int(len(downstream_xlist)/2),2)
		full_xlist = np.concatenate((upstream_xlist,downstream_xlist), axis=0)

		df_cl['norm_start'] = full_xlist[:,0]
		df_cl['norm_end'] = full_xlist[:,1]
		df_gggenes = df_gggenes.append(df_cl[['locus_tag','norm_start','norm_end']])

	return(df_gggenes)

def align_NormPositions(df_m):
	'''
	Align all positions to start at zero by taking the absolute max value for the start or end of the seed gene
	'''
	max_vals = df_m[df_m['is_seed']==1][['norm_start','norm_end']].abs().max(axis=1)
	df_max_vals = df_m[df_m['is_seed']==1][['region_id']]
	df_max_vals['max_val'] = max_vals
	df_m = df_m.merge(df_max_vals, on='region_id')
	df_m['norm_start'] = np.where(df_m['global_strand']== -1, df_m['norm_start']+df_m['max_val'],df_m['norm_start'])
	df_m['norm_end'] = np.where(df_m['global_strand']== -1, df_m['norm_end']+df_m['max_val'],df_m['norm_end'])
	return(df_m)



def plot_RepresentativeRegions(df_sr, df_cr):
	'''
	Produce the tables necessary to plot the representative regions along with associated metadata
	'''

	## format for plotting
	df_m = df_sr[df_sr['label_representative'] == df_sr['full_index_pos']][['region_id','assembly_id','contig_id','locus_tag','cds_start','cds_end','strand','global_strand','pseudo_check','pident','qcovs','evalue',
	'dbscan_label','mean_dissimilarity','representative_relative_dissimilarity','ortho_cluster_id',
	'genus','species','refseq_product','refseq_gene','new_pos','is_seed','order']]

	#reverse gene order for global_strand == -1 regions by conditionalally multplying by -1
	df_m['cds_start1'] = np.where(df_m['global_strand']==-1, df_m['cds_start']*-1, df_m['cds_start'])
	df_m['cds_end1'] = np.where(df_m['global_strand']==-1, df_m['cds_end']*-1, df_m['cds_end'])

	#convert to gggenes input format and save
	df_gggenes = convert_ToGGGenesInput(df_m=df_m,filter_on='dbscan_label')
	df_m = df_m.merge(df_gggenes,on='locus_tag')
	df_m = align_NormPositions(df_m=df_m)

	df_m.to_csv(pjoin('internal_data','rtable_region_representatives.csv'),index=None)




def get_UniqueClustersWithinCluster(df_cr, df_sr, cluster_label_id):
	'''
	'''
#troubleshooting
# os.chdir('/Users/owlex/Downloads/gg_tutorial/example_search/mexb')


# conn = sql.connect('seed_results.db')
# df_sr = pd.read_csv(pjoin('internal_data','full_region.csv'))
# df_cr = pd.read_sql_query("SELECT * from dbscan_label_representatives", conn)
# cluster_label_id = 10

	## subset for the cluster of interest
	rep_list = df_cr[df_cr['dbscan_label']==cluster_label_id]['seed_region_id'].tolist()
	df_jac = subset_SimilarityMatrix(rep_list=rep_list)
	df_jac['seed_region_id'] = rep_list
	df_mjac = pd.melt(df_jac, id_vars=['seed_region_id'])

	## group all subclusters with 0 dissimilarity
	store_unique_cwc = defaultdict(list)
	found_regions = list()
	for c in df_mjac['seed_region_id'].tolist():
		if c in found_regions:
			continue
		temp_list = df_mjac[ (df_mjac['seed_region_id'] == c) & (df_mjac['value'] == 0)]['variable'].tolist()
		store_unique_cwc[c] = temp_list
		[found_regions.append(c1) for c1 in temp_list if c1 not in found_regions]

	## extract the index of subgroup representatives and gene clusters within them to a table
	df_subind = pd.DataFrame(list(store_unique_cwc.values()), index=store_unique_cwc.keys())
	df_subind['representative_subgroup'] = df_subind.index
	df_subind = pd.melt(df_subind, id_vars=['representative_subgroup'])
	df_subind = df_subind.drop(labels=['variable'],axis=1)
	df_subind = df_subind.rename(columns={'value':'subgroup_member'})
	df_subind = df_subind.dropna()
		
	## get counts of each unique region group
	list(store_unique_cwc.keys())
	store_unique_cwc_count = list()
	for l in store_unique_cwc.keys():
		store_unique_cwc_count.append(len(store_unique_cwc[l]))
	df_cwc = pd.DataFrame({
		'seed_region_id':list(store_unique_cwc.keys()),
		'count':store_unique_cwc_count})

	# add in relative dissimilarity data, sort by ascending dissimilarity, add unique identifier, and store
	df_cwc = df_cwc.merge(df_cr[['seed_region_id','dbscan_label','representative_relative_dissimilarity']], on='seed_region_id',how='inner')
	df_cwc = df_cwc.sort_values(by=['representative_relative_dissimilarity','count'], ascending=[True,False])
	df_cwc['cwc_id'] = [i for i in range(len(df_cwc))]
	df_cwc = df_cwc.rename(columns={'seed_region_id':'region_id'},inplace=False)
	df_cwc.to_csv(pjoin('internal_data','rtable_cwc_dis_counts.csv'),index=False)

	# merge data with df_sr dataframe and store
	df_cwc = df_cwc.merge(df_sr[['region_id','assembly_id','contig_id','locus_tag','cds_start','cds_end','global_strand','strand','pseudo_check','pident','qcovs','evalue','ortho_cluster_id',
	'genus','species','refseq_product','refseq_gene','new_pos','is_seed','order']],
	left_on='region_id',right_on='region_id',how='inner')

	## convert to ggenes format and save
	#reverse gene order for global_strand == -1 regions
	df_cwc['cds_start1'] = np.where(df_cwc['global_strand']==-1, df_cwc['cds_start']*-1, df_cwc['cds_start'])
	df_cwc['cds_end1'] = np.where(df_cwc['global_strand']==-1, df_cwc['cds_end']*-1, df_cwc['cds_end'])
	#
	df_gggenes = convert_ToGGGenesInput(df_m=df_cwc,filter_on='cwc_id')
	df_cwc = df_cwc.merge(df_gggenes,on='locus_tag')
	df_cwc = align_NormPositions(df_m=df_cwc)
	df_cwc.to_csv(pjoin('internal_data','rtable_cwc_regions.csv'),index=None)

	## save a copy of subgroup data in the subgroup folder with the cluster_label_id as suffix. This is for the user to use in downstream applications
	df_subgroup_ref = df_cwc[['region_id','cwc_id']].drop_duplicates()
	df_subgroup_ref = df_subgroup_ref.merge(df_subind, left_on='region_id',right_on='representative_subgroup')
	df_subgroup_ref = df_subgroup_ref.drop(labels='region_id',axis=1)
	# rename so it's easier for the user to know what columns mean
	df_subgroup_ref = df_subgroup_ref.rename(columns={'cwc_id':'subgroup_id'})
	df_cwc = df_cwc.rename(columns={'cwc_id':'subgroup_id'})

	# if the subgroup folder doesn't exist, save it
	make_OutputDirectory(new_directory='subgroups')
	df_subgroup_ref.to_csv(pjoin('subgroups','subgroup_key_g{}.csv'.format(str(cluster_label_id))),index=None)
	df_cwc.to_csv(pjoin('subgroups','subgroup_regions_g{}.csv'.format(str(cluster_label_id))),index=None)



def retrieve_RepresentativeSeeds(tip_label_type, df_cr):
	'''
	Get the amino acid sequence of each representative seed EXCEPT for the group -1 representative
	'''
	# pd.set_option('display.max_columns', None)
	# tip_label_type = 'full'
	# os.chdir('/Users/owlex/Dropbox/Documents/Northwestern/Hartmann_Lab/syntenease_project/gtr/testbed/dataset4/test2/mexb')

	# get just the representative sequences into a new df 
	df_rep = df_cr[(df_cr['label_representative'] == df_cr['full_index_pos']) & (df_cr['dbscan_label'] != -1 )][['dbscan_label','seed_region_id']]

	# create tip_label column with specified content
	if tip_label_type == 'full':
		df_rep['tip_label'] = df_rep['seed_region_id'] + '--g' + df_cr['dbscan_label'].astype(int).astype(str)
	if tip_label_type == 'group':
		df_rep['tip_label'] = 'g'+df_rep['dbscan_label'].astype(int).astype(str)

	# write representative sequences to faa file
	with open(pjoin('internal_data','representative_seed_sequences.faa'), 'w') as outfile:
		for record in SeqIO.parse('group_region_seqs.faa', 'fasta'):
			if record.id in df_rep['seed_region_id'].tolist():
				tip_entry = df_rep[df_rep['seed_region_id'] == record.id]['tip_label'].item()
				outfile.write('>{}\n'.format(tip_entry))
				outfile.write('{}\n'.format(str(record.seq)))


def create_RepresentativeSeedTree():
	'''
	Run mafft and fasttree on the representative seed sequences
	'''
	os.system('mafft {} > {}'.format(pjoin('internal_data','representative_seed_sequences.faa'), pjoin('internal_data','representative_seed_sequences.aln')))
	os.system('FastTree {} > {}'.format(pjoin('internal_data','representative_seed_sequences.aln'), pjoin('internal_data','representative_seed_sequences.nwk')))


#------- start script ------#
def MakeVisualizations(
	UserInput_main_dir,
	UserInput_output_dir_name,
	UserInput_cluster_label_id,
	UserInput_visualization_type,
	UserInput_script_path,
	UserInput_image_format,
	UserInput_tip_label_type,
	UserInput_tip_label_size):
	'''
	'''
	change_ToWorkingDirectory(directory_name = pjoin(UserInput_main_dir,UserInput_output_dir_name))
	conn = sql.connect('seed_results.db')
	df_sr = pd.read_csv(pjoin('internal_data','full_region.csv'))
	df_cr = pd.read_sql_query("SELECT * from dbscan_label_representatives", conn)

	# main visualization
	if UserInput_visualization_type == 'main':
		print('Arranging clusters')
		plot_RepresentativeRegions(df_sr=df_sr,df_cr=df_cr)
		print('plotting in R...\n')
		os.system('Rscript {}/visualize_main_region.R {} {} {}'.format(UserInput_script_path, pjoin(os.getcwd(),'internal_data'), pjoin(os.getcwd(),'visualizations'), UserInput_image_format ))
		# check_call(['Rscript', '{}/visualize_main_region.R'.format(UserInput_script_path), '{}'.format(pjoin(os.getcwd(),'internal_data')), '{}'.format(pjoin(os.getcwd(),'visualizations')) ], stdout=STDOUT, stderr=STDOUT) #DEVNULL
		# check_call(['Rscript', 'visualize_main_region.R', '{}'.format(pjoin(os.getcwd(),'internal_data')), '{}'.format(pjoin(os.getcwd(),'visualizations')) ], stdout=STDOUT, stderr=STDOUT) #DEVNULL
		print('\nAll outputs files written to {}\n'.format(os.getcwd()))
		return

	## group visualization
	if UserInput_visualization_type == 'group':
		get_UniqueClustersWithinCluster(df_cr=df_cr, df_sr=df_sr, cluster_label_id=UserInput_cluster_label_id)
		print('plotting in R...\n')
		os.system('Rscript {}/visualize_group.R {} {} {}'.format(UserInput_script_path, pjoin(os.getcwd(),'internal_data'), pjoin(os.getcwd(),'visualizations'), UserInput_image_format ))
		return

	## tree visualization
	if UserInput_visualization_type == 'tree':
		print('Making a phylogenetic tree from representative seed sequences')
		retrieve_RepresentativeSeeds(tip_label_type=UserInput_tip_label_type, df_cr=df_cr)
		create_RepresentativeSeedTree()
		print('plotting in R...\n')
		os.system('Rscript {}/visualize_tree.R {} {} {} {}'.format(UserInput_script_path, pjoin(os.getcwd(),'internal_data'), pjoin(os.getcwd(),'visualizations'), UserInput_image_format,  UserInput_tip_label_size))
		return

#------- endscript ------#




